@@ -37,15 +37,15 @@
 #include <string.h>
 #include <sys/socket.h>
 
-#ifdef HAVE_LIBEDIT
-#  include <stdio.h>
-#  ifdef HAVE_EDIT_READLINE_READLINE_H
-#    include <edit/readline/readline.h>
-#  elif HAVE_READLINE_READLINE_H
-#    include <readline/readline.h>
-#  else
-#    include <editline/readline.h>
+#ifdef HAVE_EDIT_READLINE_READLINE_H
+#  include <edit/readline/readline.h>
+#elif HAVE_READLINE_READLINE_H
+#  include <readline/readline.h>
+#  ifdef HAVE_READLINE_HISTORY_H
+#    include <readline/history.h>
 #  endif
+#else
+#  include <editline/readline.h>
 #endif
 
 #include "vcli.h"
@@ -54,18 +54,12 @@
 #include "varnishapi.h"
 #include "vss.h"
 
-#ifdef HAVE_LIBEDIT
 #define RL_EXIT(status) \
 	do { \
 		rl_callback_handler_remove(); \
 		exit(status); \
 	} while (0)
 
-void send_line(char *l);
-
-#else
-#define RL_EXIT(status) exit(status)
-#endif
 
 static double timeout = 5;
 
@@ -173,12 +167,11 @@ do_args(int sock, int argc, char * const *argv)
 	exit(1);
 }
 
-#ifdef HAVE_LIBEDIT
 /* Callback for readline, doesn't take a private pointer, so we need
  * to have a global variable.
  */
 static int _line_sock;
-void send_line(char *l)
+static void send_line(char *l)
 {
 	if (l) {
 		cli_write(_line_sock, l);
@@ -188,7 +181,41 @@ void send_line(char *l)
 		RL_EXIT(0);
 	}
 }
-#endif
+
+static char *commands[256];
+static char *
+command_generator (const char *text, int state)
+{
+	static int list_index, len;
+	const char *name;
+
+	/* If this is a new word to complete, initialize now.  This
+	   includes saving the length of TEXT for efficiency, and
+	   initializing the index variable to 0. */
+	if (!state) {
+		list_index = 0;
+		len = strlen(text);
+	}
+
+	while ((name = commands[list_index]) != NULL) {
+		list_index++;
+		if (strncmp (name, text, len) == 0)
+			return (strdup(name));
+	}
+	/* If no names matched, then return NULL. */
+	return (NULL);
+}
+
+static char **
+varnishadm_completion (const char *text, int start, int end)
+{
+	char **matches;
+	(void)end;
+	matches = (char **)NULL;
+	if (start == 0)
+		matches = rl_completion_matches(text, command_generator);
+	return (matches);
+}
 
 /*
  * No arguments given, simply pass bytes on stdin/stdout and CLI socket
@@ -202,11 +229,6 @@ pass(int sock)
 	int i;
 	char *answer = NULL;
 	unsigned u, status;
-#ifndef HAVE_LIBEDIT
-	int n;
-#endif
-
-#ifdef HAVE_LIBEDIT
 	_line_sock = sock;
 	rl_already_prompted = 1;
 	if (isatty(0)) {
@@ -214,15 +236,46 @@ pass(int sock)
 	} else {
 		rl_callback_handler_install("", send_line);
 	}
-#endif
+	rl_attempted_completion_function = varnishadm_completion;
 
-	cli_write(sock, "banner\n");
 	fds[0].fd = sock;
 	fds[0].events = POLLIN;
 	fds[1].fd = 0;
 	fds[1].events = POLLIN;
+
+	/* Grab the commands, for completion */
+	cli_write(sock, "help\n");
+	u = VCLI_ReadResult(fds[0].fd, &status, &answer, timeout);
+	if (!u) {
+		char *t, c[128];
+		if (status == CLIS_COMMS) {
+			RL_EXIT(0);
+		}
+		t = answer;
+
+		i = 0;
+		while (*t) {
+			if (sscanf(t, "%127s", c) == 1) {
+				commands[i++] = strdup(c);
+				while (*t != '\n' && *t != '\0')
+					t++;
+				if (*t == '\n')
+					t++;
+			} else {
+				/* what? */
+				fprintf(stderr, "Unknown command '%s' parsing "
+					"help output. Tab completion may be "
+					"broken\n", t);
+				break;
+			}
+		}
+	}
+	cli_write(sock, "banner\n");
 	while (1) {
 		i = poll(fds, 2, -1);
+		if (i == -1 && errno == EINTR) {
+			continue;
+		}
 		assert(i > 0);
 		if (fds[0].revents & POLLIN) {
 			/* Get rid of the prompt, kinda hackish */
@@ -244,25 +297,10 @@ pass(int sock)
 				free(answer);
 				answer = NULL;
 			}
-#ifdef HAVE_LIBEDIT
 			rl_forced_update_display();
-#endif
 		}
 		if (fds[1].revents & POLLIN) {
-#ifdef HAVE_LIBEDIT
 			rl_callback_read_char();
-#else
-			n = read(fds[1].fd, buf, sizeof buf);
-			if (n == 0) {
-				AZ(shutdown(sock, SHUT_WR));
-				fds[1].fd = -1;
-			} else if (n < 0) {
-				RL_EXIT(0);
-			} else {
-				buf[n] = '\0';
-				cli_write(sock, buf);
-			}
-#endif
 		}
 	}
 }
